from typing import Awaitable, Callable, Dict, List, Union
from rekuest.api.schema import ProvisionFragment, WidgetInput
from rekuest.definition.registry import ActorBuilder, get_default_definition_registry
from rekuest.structures.registry import StructureRegistry
from rekuest.structures.default import get_default_structure_registry
from rekuest.actors.base import Actor


def register(
    builder: ActorBuilder = None,
    package: str = None,
    interface: str = None,
    widgets: Dict[str, WidgetInput] = None,
    interfaces: List[str] = [],
    on_provide: Callable[[ProvisionFragment], Awaitable[dict]] = None,
    on_unprovide: Callable[[], Awaitable[dict]] = None,
    structure_registry: StructureRegistry = None,
):
    """Register a function or actor to the default definition registry.

    This is a convenience function to register a function or actor to the default definition registry.
    It is equivalent to calling the register method on the default definition registry.

    Args:
        function_or_actor (Union[Callable, Actor]): The function or Actor
        builder (ActorBuilder, optional): An actor builder (see ActorBuilder). Defaults to None.
        package (str, optional): The package you want to register this function in. Defaults to standard app package    .
        interface (str, optional): The name of the function. Defaults to the functions name.
        widgets (Dict[str, WidgetInput], optional): A dictionary of parameter key and a widget. Defaults to the default widgets as registered in the structure registry .
        interfaces (List[str], optional): Interfaces that this node adheres to. Defaults to [].
        on_provide (Callable[[ProvisionFragment], Awaitable[dict]], optional): Function that shall be called on provide (in the async eventloop). Defaults to None.
        on_unprovide (Callable[[], Awaitable[dict]], optional): Function that shall be called on unprovide (in the async eventloop). Defaults to None.
        structure_registry (StructureRegistry, optional): The structure registry to use for this Actor (used to shrink and expand inputs). Defaults to None.
    """

    def real_decorator(function):
        # Simple bypass for now
        def wrapped_function(*args, **kwargs):
            return function(*args, **kwargs)

        get_default_definition_registry().register(
            function_or_actor=function,
            builder=builder,
            package=package,
            interface=interface,
            widgets=widgets or dict(),
            interfaces=interfaces,
            on_provide=on_provide,
            on_unprovide=on_unprovide,
            structure_registry=structure_registry or get_default_structure_registry(),
        )

    return real_decorator



from arkitekt.apps.connected import ConnectedApp


class Arkitekt(ConnectedApp):
    """Arkitekt

    An app that connected to the services of the arkitekt Api,
    it comes included with the following services:

    - Rekuest: A service for that handles requests to the arkitekt Api as well as provides an interface to provide functionality on the arkitekt Api.
    - Herre: A service for that handles the authentication and authorization of the user
    - Fakts: A service for that handles the discovery and retrieval of the configuration of the arkitekt Api
    - Mikro: A service for that handles the storage and data of microscopy data

    Apps have to be always used within a context manager, this is to ensure that the services are properly closed when the app is no longer needed.

    Example:
        >>> from arkitekt import Arkitekt
        >>> app = Arkitekt()
        >>> with app:
        >>>     # Do stuff
        >>> # App is closed



    """




"""
This modules provides the main app. It is responsible for setting up the connection to the mikro-server and
handling authentification and setting up the configuration. Mikro handles the creation of the datalayer and
the graphql client.

You can compose this app together with other apps to create a full fledged app. (Like combining with
arkitekt to enable to call functions that you define on the app)

Example:

    A simple app that takes it configuraiton from basic.fakts and connects to the mikro-server.
    You can define all of the logic within the context manager

    ```python
    from mikro.app import MikroApp

    app = MikroApp(fakts=Fakts(subapp="basic"))

    with app:
        # do stuff

    ```

    Async Usage:


    ```python
    from mikro.app import MikroApp

    app = MikroApp(fakts=Fakts(subapp="basic"))

    async with app:
        # do stuff

    ```

 
"""


from arkitekt.apps.herre import HerreApp
from pydantic import Field
from mikro.datalayer import DataLayer
from mikro.mikro import Mikro
from mikro.rath import MikroLinkComposition, MikroRath
from rath.links.split import SplitLink
from rath.contrib.fakts.links.aiohttp import FaktsAIOHttpLink
from rath.contrib.fakts.links.websocket import FaktsWebsocketLink
from rath.contrib.herre.links.auth import HerreAuthLink
from mikro.contrib.fakts.datalayer import FaktsDataLayer
from graphql import OperationType


class ArkitektMikro(Mikro):
    rath: MikroRath = Field(
        default_factory=lambda: MikroRath(
            link=MikroLinkComposition(
                auth=HerreAuthLink(),
                split=SplitLink(
                    left=FaktsAIOHttpLink(fakts_group="mikro"),
                    right=FaktsWebsocketLink(fakts_group="mikro"),
                    split=lambda o: o.node.operation != OperationType.SUBSCRIPTION,
                ),
            )
        )
    )
    datalayer: DataLayer = Field(
        default_factory=lambda: FaktsDataLayer(fakts_group="mikro.datalayer")
    )


class MikroApp(HerreApp):
    """Mikro App

    It is responsible for setting up the connection to the mikro-server and
    handling authentification and setting up the configuration. Mikro handles the creation of the datalayer and
    the graphql client.

    You can compose this app together with other apps to create a full fledged app. (Like combining with
    arkitekt to enable to call functions that you define on the app). See the example in the docstring.

    Attributes:
        fakts (Fakts): The fakts instance to use.
        mikro (Mikro): The mikro instance to use.
        herre (Herre): The herre instance to use.

    """

    mikro: ArkitektMikro = Field(default_factory=ArkitektMikro)



import argparse
from enum import Enum
from arkitekt.cli.prod.run import import_directory_and_start
from fakts import Fakts
from fakts.discovery.static import StaticDiscovery
from herre import Herre
from rich.console import Console
from rich.prompt import Prompt, Confirm
import os
import asyncio

try:
    from rich.traceback import install

    install()
except:
    pass

directory = os.getcwd()


class ArkitektOptions(str, Enum):
    INIT = "init"
    DEV = "dev"
    LOGOUT = "logout"
    LOGIN = "login"
    RUN = "run"
    WAIT = "wait"


agent_script = f'''
from arkitekt import register

@register()
def test_show(name: str)-> str:
    """Demo Node

    This demo node will just pass through
    a name 

    Args:
        name (str): The name

    Returns:
        str: The pass throughed Name
    """
    return name


'''

mikro_script = f"""
from arkitekt import Arkitekt

app = Arkitekt()

with app:
    # Implement your logig here
"""

HEADER = """
                __    _  __         __    __ 
  ____ _ _____ / /__ (_)/ /_ ___   / /__ / /_
 / __ `// ___// //_// // __// _ \ / //_// __/
/ /_/ // /   / ,<  / // /_ /  __// ,<  / /_  
\__,_//_/   /_/|_|/_/ \__/ \___//_/|_| \__/  
                                             
"""


template_map = {"agent": agent_script, "client": mikro_script}


def main(
    script=ArkitektOptions.INIT,
    name=None,
    path=".",
    refresh=False,
    scan_network=None,
    beacon=None,
    template=None,
    services=[],
):

    console = Console()
    console.print(HEADER)

    if path == ".":
        app_directory = os.getcwd()
        name = name or os.path.basename(app_directory)
    else:
        app_directory = os.path.join(os.getcwd(), path)
        name = name or path

    fakts_path = os.path.join(app_directory, "fakts.yaml")
    run_script_path = os.path.join(app_directory, "run.py")

    if script == ArkitektOptions.DEV:

        from arkitekt.cli.dev.autostage import watch_directory_and_stage

        if not os.path.isfile(run_script_path):
            console.print(f"Could't find a run.py in {app_directory} {run_script_path}")
            return
        if not os.path.isfile(fakts_path):
            console.print(f"{app_directory} does not have a valid fakts.yaml")
            return

        asyncio.run(watch_directory_and_stage(path, entrypoint="run"))

    if script == ArkitektOptions.RUN:

        if not os.path.isfile(run_script_path):
            console.print(f"Could't find a run.py in {app_directory} {run_script_path}")
            return
        if not os.path.isfile(fakts_path):
            console.print(f"{app_directory} does not have a valid fakts.yaml")
            return

        asyncio.run(import_directory_and_start(path, entrypoint="run"))

    if script == ArkitektOptions.LOGIN:

        if not os.path.isfile(fakts_path):
            console.print(
                f"Directory does not containt a valid fakts.yaml. Please initialize through 'arkitekt init {path}' first"
            )
            return

        fakts = Fakts(grants=[], fakts_path=fakts_path)
        if not fakts.loaded:
            console.print(
                f"Configuration in fakts.yaml is not sufficient please reinitilize through 'arkitekt init {path}'"
            )

        herre = Herre(fakts=fakts)
        console.print("Loggin in")
        herre.login()

    if script == ArkitektOptions.INIT:

        if os.path.exists(app_directory):
            console.print(f"App director already existed")
        else:
            os.mkdir(app_directory)
            console.print(f"Created a new App Directory at {app_directory}")

        initialize_fakts = True
        if os.path.isfile(fakts_path):
            console.print(
                f"There seems to have been a previous app initialized at {app_directory}"
            )
            if refresh is None:
                initialize_fakts = Confirm.ask(
                    "Do you want to refresh the Apps configuration? "
                )
                if not initialize_fakts:
                    console.print(f"Nothing done. Bye :smiley: ")
                    return
            else:
                initialize_fakts = refresh

        if initialize_fakts:

            console.print(f"Initializing a new Configuration for {name}")

            if scan_network is None:
                scan_network = Confirm.ask(
                    "Do you want to automatically scan the network for local instances?"
                )
                from fakts.grants.remote.device_code import DeviceCodeGrant
                from fakts.discovery.advertised import AdvertisedDiscovery

                if scan_network:
                    discovery = AdvertisedDiscovery()
                else:
                    url = Prompt.ask(
                        "Give us the path to your fakts instance",
                        default="http://localhost:8000/f/",
                    )
                    discovery = StaticDiscovery(base_url=url)

                has_webrowser = Confirm.ask(
                    "Do you have a webbrowser installed? Otherwise we will generate a code that you need to enter manually"
                )

                fakts = Fakts(
                    grant=DeviceCodeGrant(
                        discovery=discovery, open_browser=has_webrowser
                    ),
                    fakts_path=fakts_path,
                    force_refresh=True,
                )

                with fakts as f:
                    f.load()

        save_run = True
        if os.path.isfile(run_script_path):
            console.print("Initializing Templates for the App")
            save_run = Confirm.ask(
                "run.py already exists? Do you want to overwrite the run.py file? :smiley: "
            )

        if save_run:
            if not template:
                template = Prompt.ask(
                    "Which Template do you want to use?",
                    choices=list(template_map.keys()),
                    default="agent",
                )

            with open(run_script_path, "w") as f:
                f.write(template_map[template])
                console.print("Create new Entrypoint for this App")


def entrypoint():
    parser = argparse.ArgumentParser(description="Say hello")
    parser.add_argument("script", type=ArkitektOptions, help="The Script Type")
    parser.add_argument("path", type=str, help="The Path", nargs="?", default=".")
    parser.add_argument("--name", type=str, help="The Name of this script")
    parser.add_argument("--refresh", type=bool, help="Do you want to refresh")
    parser.add_argument("--scan-network", type=bool, help="Do you want to refresh")
    parser.add_argument("--beacon", type=str, help="The adress of the beacon")
    parser.add_argument("--template", type=str, help="The run script template")
    parser.add_argument(
        "--services",
        type=str,
        help="The services you want to connect to (seperated by ,)",
    )
    args = parser.parse_args()

    main(
        script=args.script,
        path=args.path,
        name=args.name,
        refresh=args.refresh,
        scan_network=args.scan_network,
        template=args.template,
        services=args.services.split(",") if args.services else [],
    )


if __name__ == "__main__":
    entrypoint()



from datetime import datetime
from importlib import reload, import_module
import asyncio
import asyncio
import sys
import time
from arkitekt import Arkitekt
from rich.console import Console


try:
    from watchdog.observers import Observer
    from watchdog.events import (
        FileModifiedEvent,
        FileSystemEventHandler,
    )
except ImportError:
    print("Please install watchdog to use this feature")
    sys.exit(1)

try:

    import janus
except ImportError:
    print("Please install janus to use this feature")
    sys.exit(1)

import os
import threading

from rekuest.definition.registry import (
    get_default_definition_registry,
)


class QueueHandler(FileSystemEventHandler):
    def __init__(self, *args, sync_q=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.queue = sync_q

    def on_any_event(self, event):
        self.queue.put(event)
        self.queue.join()


def watcher(path, queue, event: threading.Event):
    event_handler = QueueHandler(sync_q=queue)
    observer = Observer()
    observer.schedule(event_handler, path, recursive=True)
    observer.start()

    while not event.is_set():
        time.sleep(1)

    print("Cancelled because threading event is set")


async def buffered_queue(queue, timeout=4):
    buffer = [None]

    async def iterator():
        while True:
            nana = await queue.get()
            queue.task_done()
            if isinstance(nana, FileModifiedEvent):
                if nana.src_path.lower().endswith((".py")):
                    buffer.append(nana)

    loop = asyncio.get_event_loop()
    loop.create_task(iterator())
    while True:
        await asyncio.sleep(timeout)
        yield buffer[len(buffer) - 1]
        buffer = [None]


class Host:
    def __init__(
        self,
        path=None,
        entrypoint="run",
    ) -> None:

        if path == ".":
            self.watch_path = os.getcwd()
            sys.path.insert(0, self.watch_path)
        else:
            self.watch_path = os.path.join(os.getcwd(), path)

        self.module_path = f"{path}.{entrypoint}" if path != "." else f"{entrypoint}"
        self.module = import_module(self.module_path)
        self.provide_task = None
        self.console = Console()
        os.environ["ARKITEKT_AGENT_DEBUG"] = "True"

    async def reprovide(self):
        registry = get_default_definition_registry()
        registry.reset()
        reload(self.module)

        try:
            if registry.has_definitions():
                app = Arkitekt()

                async with app:
                    self.console.print(
                        f"[bold green] --------------- Providing {datetime.now()} ------------- [/bold green]"
                    )
                    await app.rekuest.run()

        except Exception as e:
            self.console.print(
                f"[bold red] --------------- Error in App {datetime.now()} ------------- [/bold red]"
            )
            self.console.print_exception()

        except asyncio.CancelledError as e:
            self.console.print(
                f"[bold yellow] --------------- App Cancelled {datetime.now()} ------------- [/bold yellow]"
            )
            raise e

    async def restart(self):
        loop = asyncio.get_event_loop()

        if self.provide_task:
            self.provide_task.cancel()
            try:
                await self.provide_task
            except asyncio.CancelledError as e:
                pass

        self.provide_task = loop.create_task(self.reprovide())

    async def run(self):
        await self.restart()

        jqueue = janus.Queue()

        cancel_event = threading.Event()
        loop = asyncio.get_running_loop()
        fut = loop.run_in_executor(
            None, watcher, self.watch_path, jqueue.sync_q, cancel_event
        )

        try:
            async for event in buffered_queue(jqueue.async_q, timeout=1):
                if event:
                    self.console.print(
                        f"[bold yellow] --------------- File Changed {datetime.now()} ------------- [/bold yellow]"
                    )
                    await self.restart()
        except asyncio.CancelledError as e:
            cancel_event.set()
            await fut

        jqueue.close()
        await jqueue.wait_closed()


async def watch_directory_and_stage(path="watch", entrypoint="run"):
    host = Host(path=path)
    await host.run()



from fakts.fakts import Fakts
from koil.composition.base import Composition
from pydantic import Field


class FaktsApp(Composition):
    fakts: Fakts = Field(default_factory=Fakts)
    """The fakts layer that is used for configuration
    """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['arkitekt',
 'arkitekt.apps',
 'arkitekt.cli',
 'arkitekt.cli.dev',
 'arkitekt.cli.prod',
 'arkitekt.qt']

package_data = \
{'': ['*'], 'arkitekt.qt': ['assets/dark/*', 'assets/light/*']}

install_requires = \
['fakts==0.2.10',
 'herre>=0.2.7,<0.3.0',
 'mikro==0.2.23',
 'rath==0.2.9',
 'rekuest==0.0.16']

entry_points = \
{'console_scripts': ['arkitekt = arkitekt.cli.main:entrypoint']}

setup_kwargs = {
    'name': 'arkitekt',
    'version': '0.3.6',
    'description': 'rpc and node backbone',
    'long_description': 'None',
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from arkitekt.apps.fakts import FaktsApp
from herre.fakts.herre import FaktsHerre
from herre.herre import Herre
from pydantic import Field


class HerreApp(FaktsApp):
    herre: Herre = Field(default_factory=FaktsHerre)
    """The fakts layer that is used for configuration
    """
